"""Setup tests."""

import os


def pytest_runtest_setup(item):
    """Before test runs, make sure no MlFlow ENV variable is set.

    Parameters
    ----------
    item : test function
    """
    # pylint: disable=unused-argument,protected-access
    if "MLFLOW_RUN_ID" in os.environ:
        del os.environ["MLFLOW_RUN_ID"]
    if "MLFLOW_TRACKING_URI" in os.environ:
        del os.environ["MLFLOW_TRACKING_URI"]



"""MlFlow Launcher."""

from pathlib import Path
from typing import Any, Iterable
import logging
import os
import re
import tempfile

import fromconfig
import mlflow


LOGGER = logging.getLogger(__name__)


_RUN_ID_ENV_VAR = "MLFLOW_RUN_ID"
_TRACKING_URI_ENV_VAR = "MLFLOW_TRACKING_URI"


class MlFlowLauncher(fromconfig.launcher.Launcher):
    """MlFlow Launcher.

    To configure MlFlow, add a `mlflow` entry to your config and set the
    following parameters

    - `run_id`: if you wish to restart an existing run
    - `run_name`: if you wish to give a name to your new run
    - `tracking_uri`: to configure the tracking remote
    - `experiment_name`: to use a different experiment than the custom
      experiment
    - `artifact_location`: the location of the artifacts (config files)

    Attributes
    ----------
    log_artifacts : bool, optional
        If True, save config and command as artifacts.
    log_params : bool, optional
        If True, log flattened config as parameters.
    path_command : str, optional
        Name for the command file
    path_config : str, optional
        Name for the config file.
    set_env_vars : bool, optional
        If True, set MlFlow environment variables.
    set_run_id : bool, optional
        If True, the run_id is overridden in the config.
    ignore_keys : Iterable[str], optional
        If given, don't log some parameters that have some substrings.
    include_keys : Iterable[str], optional
        If given, only log some parameters that have some substrings.
        Also shorten the flattened parameter to start at the first
        match. For example, if the config is `{"foo": {"bar": 1}}` and
        `include_keys=("bar",)`, then the logged parameter will be
        `"bar"`.
    """

    def __init__(
        self,
        launcher: fromconfig.launcher.Launcher,
        log_artifacts: bool = True,
        log_params: bool = True,
        path_command: str = "launch.sh",
        path_config: str = "config.yaml",
        set_env_vars: bool = False,
        set_run_id: bool = True,
        ignore_keys: Iterable[str] = None,
        include_keys: Iterable[str] = None,
    ):
        super().__init__(launcher=launcher)
        self.ignore_keys = ignore_keys
        self.include_keys = include_keys
        self.log_artifacts = log_artifacts
        self.log_params = log_params
        self.path_command = path_command
        self.path_config = path_config
        self.set_env_vars = set_env_vars
        self.set_run_id = set_run_id

    def __call__(self, config: Any, command: str = ""):
        if mlflow.active_run() is not None:
            print(f"Active run found: {get_url(mlflow.active_run())}")
            self.log_and_launch(config=config, command=command)
        else:
            # Create run from params in config
            params = config.get("mlflow") or {}
            run_id = params.get("run_id")
            run_name = params.get("run_name")
            tracking_uri = params.get("tracking_uri")
            experiment_name = params.get("experiment_name")
            artifact_location = params.get("artifact_location")

            # Setup experiment and general MlFlow parameters
            if tracking_uri is not None:
                mlflow.set_tracking_uri(tracking_uri)
            if experiment_name is not None:
                if mlflow.get_experiment_by_name(experiment_name) is None:
                    mlflow.create_experiment(experiment_name, artifact_location=artifact_location)
                mlflow.set_experiment(experiment_name)

            # Start MlFlow run, log information and launch
            with mlflow.start_run(run_id=run_id, run_name=run_name) as run:
                print(f"Started run: {get_url(run)}")
                self.log_and_launch(config=config, command=command)

    def log_and_launch(self, config: Any, command: str = ""):
        """Log and launch config

        Parameters
        ----------
        config : Any
            Config
        command : str, optional
            Command
        """
        # Log artifacts
        if self.log_artifacts:
            LOGGER.info(f"Logging artifacts {self.path_config} and {self.path_command}")
            dir_artifacts = tempfile.mkdtemp()
            fromconfig.dump(config, Path(dir_artifacts, self.path_config))
            with Path(dir_artifacts, self.path_command).open("w") as file:
                file.write(f"fromconfig {self.path_config} - {command}")
            mlflow.log_artifacts(local_dir=dir_artifacts)

        # Log parameters by batches of 100
        if self.log_params:
            LOGGER.info("Logging params")
            params = get_params(config, self.ignore_keys, self.include_keys)
            for idx in range(0, len(params), 100):
                mlflow.log_params(dict(params[idx : idx + 100]))

        # A bit risky as ENV variables are global, risk conflicts with
        # another place that would set / use these
        if self.set_env_vars:
            LOGGER.info(f"Setting ENV variables {_RUN_ID_ENV_VAR} and {_TRACKING_URI_ENV_VAR}")
            os.environ[_RUN_ID_ENV_VAR] = mlflow.active_run().info.run_id
            os.environ[_TRACKING_URI_ENV_VAR] = mlflow.tracking.get_tracking_uri()

        # Update run_id to override config for future launches
        if self.set_run_id:
            LOGGER.info("Setting mlflow.run_id in config")
            run_id = mlflow.active_run().info.run_id
            config = fromconfig.utils.merge_dict(config, {"mlflow": {"run_id": run_id}})

        # Launch
        self.launcher(config=config, command=command)

        # Clean up the environment variables once not needed
        if self.set_env_vars:
            LOGGER.info(f"Cleaning ENV variables {_RUN_ID_ENV_VAR} and {_TRACKING_URI_ENV_VAR}")
            del os.environ[_RUN_ID_ENV_VAR]
            del os.environ[_TRACKING_URI_ENV_VAR]


# log_params only accepts alphanumerics, period, space, dash, underscore
_FORBIDDEN = re.compile(r"[^0-9a-zA-Z_\. \-/]+")


def get_params(config, ignore_keys=None, include_keys=None):
    """Log param if coherent with ignore keys and include keys."""
    params = []
    for key, value in fromconfig.utils.flatten(config):
        if include_keys and not any(k in key for k in include_keys):
            continue
        if include_keys:
            for k in include_keys:
                index = key.find(k)
                if index != -1:
                    key = key[index:]
        if ignore_keys and any(k in key for k in ignore_keys):
            continue
        params.append((_FORBIDDEN.sub("_", key), value))
    return params


def get_url(run) -> str:
    return f"{mlflow.get_tracking_uri()}/experiments/{run.info.experiment_id}/runs/{run.info.run_id}"



"""Test launcher."""

import collections
import functools

import pytest
import fromconfig

import fromconfig_mlflow
from fromconfig_mlflow.launcher import get_params


def test_launcher_is_discovered():
    """Test that the Launcher is discovered as fromconfig extension."""
    fromconfig.utils.testing.assert_launcher_is_discovered("mlflow", fromconfig_mlflow.MlFlowLauncher)


@pytest.mark.parametrize(
    "launcher, params, expected, counts",
    [
        pytest.param(
            {"log": "mlflow"},
            {},
            {"start_run": {"kwargs": {"run_name": None}}},
            {"start_run": 1, "log_artifacts": 1},
            id="default",
        ),
        pytest.param(
            {"log": "mlflow"},
            {"experiment_name": "test"},
            {"start_run": {"kwargs": {"run_name": None}}},
            {"start_run": 1, "log_artifacts": 1},
            id="experiment_name",
        ),
        pytest.param(
            {"log": "mlflow"},
            {"run_name": "test"},
            {"start_run": {"kwargs": {"run_name": "test"}}},
            {"start_run": 1, "log_artifacts": 1},
            id="run_name",
        ),
        pytest.param(
            {"log": ["mlflow", "mlflow"]},
            {"run_name": "test"},
            {"start_run": {"kwargs": {"run_name": "test"}}},
            {"start_run": 1, "log_artifacts": 2},
            id="duplicate",
        ),
        pytest.param(
            {
                "log": [
                    {"_attr_": "fromconfig_mlflow.MlFlowLauncher", "log_params": False},
                    {"_attr_": "fromconfig_mlflow.MlFlowLauncher", "log_params": True},
                ]
            },
            {"run_name": "test"},
            {"start_run": {"kwargs": {"run_name": "test"}}},
            {"start_run": 1, "log_artifacts": 2, "log_params": 1},
            id="duplicate-with-launches-param",
        ),
        pytest.param(
            {
                "log": [
                    {"_attr_": "fromconfig_mlflow.MlFlowLauncher", "log_artifacts": False, "set_env_vars": True},
                    {"_attr_": "fromconfig_mlflow.MlFlowLauncher", "log_artifacts": True},
                ]
            },
            {"run_name": "test"},
            {"start_run": {"kwargs": {"run_name": "test"}}},
            {"start_run": 1, "log_artifacts": 1, "log_params": 2},
            id="no-duplicate-log-artifacts",
        ),
    ],
)
def test_launcher(launcher, params, expected, counts, monkeypatch):
    """Test cli.main."""
    import mlflow  # pylint: disable=import-outside-toplevel

    # Setup monkey patching to track MlFlow calls
    got = {}
    counter = collections.Counter()
    names = ["log_artifacts", "log_params", "start_run"]
    for name in names:
        func = getattr(mlflow, name)

        def _func(*args, f=None, n=None, **kwargs):
            got.update({n: {"args": args, "kwargs": kwargs}})
            counter.update([n])
            return f(*args, **kwargs)

        monkeypatch.setattr(mlflow, name, functools.partial(_func, n=name, f=func))

    # Create config, launcher, and launch
    config = {"run": None, "launcher": launcher, "mlflow": params}
    launcher = fromconfig.launcher.DefaultLauncher.fromconfig(config.pop("launcher"))
    launcher(config, "run")

    def _check_expected(left, right):
        if isinstance(left, dict):
            for key, value in left.items():
                _check_expected(value, right[key])
        else:
            assert left == right

    # Check that all MlFlow functions were called
    assert all(name in got for name in names)

    # Check that number of calls matches expectations
    assert all(counter[key] == count for key, count in counts.items()), counter

    # Check that kwargs / args match expectations
    _check_expected(expected, got)


@pytest.mark.parametrize(
    "params, ignore_keys, include_keys, expected",
    [
        pytest.param({"foo": 0}, None, None, [("foo", 0)], id="default"),
        pytest.param({"foo": 0, "bar": 1}, None, None, [("foo", 0), ("bar", 1)], id="default-multi"),
        pytest.param({"foo": 0}, (), (), [("foo", 0)], id="default-empty-tuple"),
        pytest.param({"foo": 0}, (), ("foo",), [("foo", 0)], id="include"),
        pytest.param({"foo": 0, "bar": 1}, (), ("foo",), [("foo", 0)], id="include-multi"),
        pytest.param({"foo": 0}, ("foo",), (), [], id="ignore"),
        pytest.param({"foo": 0, "bar": 1}, ("foo",), (), [("bar", 1)], id="ignore-multi"),
        pytest.param(
            {"foo": {"bar": {"bar": 0}}, "bar": 1}, (), ("bar",), [("bar.bar", 0), ("bar", 1)], id="include-conflict"
        ),
    ],
)
def test_get_params(params, ignore_keys, include_keys, expected):
    """Test log param."""
    assert get_params(params, ignore_keys, include_keys) == expected



"""Test version."""

from fromconfig_mlflow import version


def test_version():
    """Test version."""
    assert len(version.__version__.split(".")) == 3
    assert isinstance(version.__author__, str)



"""Setup script."""

from pathlib import Path
import re
import setuptools


if __name__ == "__main__":
    # Read metadata from version.py
    with Path("fromconfig_mlflow/version.py").open(encoding="utf-8") as file:
        metadata = dict(re.findall(r'__([a-z]+)__\s*=\s*"([^"]+)"', file.read()))

    # Read description from README
    with Path(Path(__file__).parent, "README.md").open(encoding="utf-8") as file:
        long_description = file.read()

    # Run setup
    setuptools.setup(
        author=metadata["author"],
        version=metadata["version"],
        classifiers=[
            "License :: OSI Approved :: Apache Software License",
            "Operating System :: OS Independent",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.6",
            "Intended Audience :: Developers",
        ],
        data_files=[(".", ["requirements.txt", "README.md"])],
        dependency_links=[],
        description=long_description.split("\n")[0],
        entry_points={"fromconfig0": ["mlflow = fromconfig_mlflow"]},
        install_requires=["fromconfig>=0.3", "mlflow"],
        long_description=long_description,
        long_description_content_type="text/markdown",
        name="fromconfig_mlflow",
        packages=setuptools.find_packages(),
        tests_require=["pytest"],
        url="https://github.com/criteo/fromconfig-mlflow",
    )
